import time
import math
import numpy
from joblib import Parallel, delayed

#from pysam.libcalignmentfile cimport AlignmentFile, AlignedSegment
from pysam import AlignmentFile, AlignedSegment



def percentile(a, q):
	size = len(a)
	percentiles=[]

	sorted_a=sorted(a)

	for v in q:
		if not size:
			percentiles.append(0)
		else:
			percentiles.append( sorted_a[ int(math.ceil((size * v) / 100.0)) - 1 ] )

	return(percentiles)


def scoring(scoring_dict,percentiles):
	score=[0]
	if scoring_dict["n_contigs"]:
		score.append(50)

	if scoring_dict["n_discordants"]:
		score.append(0)
		for p in percentiles["FA"]:
			if scoring_dict["n_discordants"]/(scoring_dict["refFA"]+scoring_dict["n_discordants"]) >= p:
				score[-1]+=5

		score.append(0)
		for p in percentiles["FB"]:
			if scoring_dict["n_discordants"]/(scoring_dict["refFB"]+scoring_dict["n_discordants"]) >= p:
				score[-1]+=5


	if scoring_dict["n_splits"]:
		score.append(0)
		for p in percentiles["RA"]:
			if scoring_dict["n_splits"]/(scoring_dict["refRA"]+scoring_dict["n_splits"]) >= p:
				score[-1]+=5

		score.append(0)
		for p in percentiles["RB"]:
			if scoring_dict["n_splits"]/(scoring_dict["refRB"]+scoring_dict["n_splits"]) >= p:
				score[-1]+=5

	return(max(score))

def get_region(samfile,str chr,int start,int end,int bp,int min_q,int max_ins, contig_number):

	cdef int low_q=0
	cdef int n_reads=0
	cdef long bases=0
	cdef int n_discs=0
	cdef int n_splits=0


	cdef int crossing_r=0
	cdef int crossing_f=0

	bam_header=samfile.header
	contig_length=bam_header["SQ"][contig_number[chr]]["LN"]

	q_start=start
	q_end=end+max_ins

	if q_end > contig_length:
		q_end=contig_length

	if q_start >= q_end:
		q_start=q_end-10

	cdef long read_reference_start
	cdef long read_reference_end

	cdef long r_start
	cdef long r_end

	for read in samfile.fetch(chr, q_start, q_end):
		if read.is_unmapped:
			continue

		read_reference_start=read.reference_start

		if not read.mate_is_unmapped:
			if read.next_reference_start > end and read_reference_start > end:
				continue
		else:
			if read_reference_start > end:
				continue

		if read.is_duplicate:
			continue

		if not (read_reference_start > end):
			n_reads+=1
			if read.mapq < min_q:
				low_q+=1

		if read.mapq < min_q:
			continue

		read_reference_end=read.reference_end
		read_reference_name=read.reference_name
		read_next_reference_name=read.next_reference_name

		r_start=read_reference_start
		r_end=read_reference_end

		if read_reference_start < bp-20 and r_end > bp+20:
			crossing_r+=1

		mate_bp_read= (read.next_reference_start < bp-50 and r_end > bp+50)
		discordant= ( abs(read.isize) > max_ins or read_next_reference_name != read_reference_name )

		if mate_bp_read and not discordant:
			crossing_f+=1

		if read_reference_end < start:
			continue
		elif read_reference_start > end:
			continue

		if read_reference_start < start:
			r_start=start

		if read_reference_end > end:
			r_end=end

		bases+=r_end-r_start+1

		if read.has_tag("SA"):
			n_splits+=1

		if discordant:
			n_discs+=1

	coverage= bases/(end-start+1)

	if n_reads > 0:
		frac_low_q=low_q/float(n_reads)
	else:
		frac_low_q=0

	return(coverage,frac_low_q,n_discs,n_splits,crossing_f,crossing_r)

def find_sv_type(chrA,chrB,inverted,non_inverted,args,sample_data,samples,library):
	if chrA != chrB:
		return("BND",".")

	p=library["contig_ploidy_{}".format(chrA)]
	for sample in samples:
		cn=int(round(sample_data[sample]["covM"]*p/library[ "avg_coverage_{}".format(chrA) ]))

	#mitochondria or similar
	if p > args.n*10:
		if cn > p*1.05:
			if inverted:
				return("DUP:INV",cn)
			else:
				return("DUP:TANDEM",cn)
		elif cn < p*0.95:
			return("DEL",cn)
		elif inverted > non_inverted:
			return("INV",cn)
		else:
			return("BND",cn)
	else:
		if cn > p:
			if inverted:
				return("DUP:INV",cn)
			else:
				return("DUP:TANDEM",cn)
		elif cn < p:
			return("DEL",cn)

		elif inverted > non_inverted:
			return("INV",cn)
		else:
			return("BND",cn)

def sv_filter(sample_data,args,chrA,chrB,posA,posB,max_ins_len,n_discordants,n_splits,library,n_discs_regionA,n_discs_regionB,n_splits_regionA,n_splits_regionB,n_contigs):
	filt="PASS"
	for sample in sample_data:

		#filter SV in high coverage regions
		if sample_data[sample]["covA"] > args.max_coverage*library[ "avg_coverage_{}".format(chrA) ]:
			return("UnexpectedCoverage")
		if sample_data[sample]["covB"] > args.max_coverage*library[ "avg_coverage_{}".format(chrB) ]:
			return("UnexpectedCoverage")
		if sample_data[sample]["covM"] > args.max_coverage*library[ "avg_coverage_{}".format(chrB) ]:
			return("UnexpectedCoverage")

		#if abs(posA-posB) > max_ins_len:
		if not n_contigs:
			if (n_discordants < args.p_ratio * sample_data[sample]["refFA"] or n_discordants < args.p_ratio * sample_data[sample]["refFB"]) and (n_splits < args.r_ratio * sample_data[sample]["refRA"] or n_splits < args.r_ratio * sample_data[sample]["refRB"]):
				return("BelowExpectedLinks")

			if (n_discordants < args.p_ratio * sample_data[sample]["covA"] or n_discordants < args.p_ratio * sample_data[sample]["covB"]) and (n_splits < args.r_ratio * sample_data[sample]["covA"] or n_splits < args.r_ratio * sample_data[sample]["covB"]):
				return("BelowExpectedLinks")

			if n_discordants > n_splits:
				if n_discordants < 0.25*n_discs_regionA or n_discordants < 0.25*n_discs_regionB:
					return("FewLinks")

			else:
				if n_splits < 0.25*n_splits_regionA or n_splits < 0.25*n_splits_regionB:
					return("FewLinks")

			if sample_data[sample]["QA"] > 0.2 or sample_data[sample]["QB"] > 0.2:
				return("RegionalQ")

			if n_discordants == 0 and (chrA != chrB):
				return("SplitsVSDiscs")

		#interchromsomal translocation, supported only by contigs
		if n_contigs and (chrA != chrB):
			if n_discordants < args.p:
				return("BelowExpectedLinks")

		#large variant, supported only by contigs but not discordant pairs
		elif n_contigs and (chrA == chrB and max_ins_len*3 < abs(posB-posA) ):
			if n_discordants < args.p:
				return("BelowExpectedLinks")

	return(filt)

def define_variant(str chrA, str bam_file_name,dict sv_clusters,args,dict library,int min_mapq,samples,dict coverage_data,contig_number,max_ins_len,contig_seqs):

	samfile  = AlignmentFile(bam_file_name, "r",reference_filename=args.ref,index_filename="{}_tiddit/{}.csi".format(args.o,samples[0]))
	variants=[]

	var_n=0
	for chrB in sv_clusters[chrA]:
		for cluster in sv_clusters[chrA][chrB]:
			n_discordants=sv_clusters[chrA][chrB][cluster]["N_discordants"]
			n_splits=sv_clusters[chrA][chrB][cluster]["N_splits"]
			n_contigs=sv_clusters[chrA][chrB][cluster]["N_contigs"]

			if (n_discordants < args.p and n_splits < args.r) and not n_contigs:
					continue

			posA=sv_clusters[chrA][chrB][cluster]["posA"]
			posB=sv_clusters[chrA][chrB][cluster]["posB"]

			if chrA == chrB and posA > posB:
				posT=posA
				posA=posB
				posB=posT

			if chrA == chrB and abs(posA-posB) < args.z:
				continue

			s=int(math.floor(sv_clusters[chrA][chrB][cluster]["startA"]/50.0))
			e=int(math.floor(sv_clusters[chrA][chrB][cluster]["endA"]/50.0))+1
			avg_a=numpy.average(coverage_data[chrA][s:e])

			if avg_a > args.max_coverage*library[ "avg_coverage_{}".format(chrA) ]:
				continue
			elif (args.max_coverage*n_discordants/avg_a < args.p_ratio/2 and args.max_coverage*n_splits/avg_a < args.r_ratio/2) and not n_contigs:
				continue

			avg_b=numpy.average(coverage_data[chrA][s:e])
			if avg_b == 0:
				continue
			elif avg_b > args.max_coverage*library[ "avg_coverage_{}".format(chrB) ]:
				continue
			elif (args.max_coverage*n_discordants/avg_b < args.p_ratio/2 and args.max_coverage*n_splits/avg_b < args.r_ratio/2) and not n_contigs:
				continue


			var_n+=1
			sample_data={}
			for sample in samples:

				coverageA,frac_low_qA,n_discsA,n_splitsA,crossing_f_A,crossing_r_A=get_region(samfile,chrA,sv_clusters[chrA][chrB][cluster]["startA"],sv_clusters[chrA][chrB][cluster]["endA"],posA,min_mapq,max_ins_len,contig_number)
				coverageB,frac_low_qB,n_discsB,n_splitsB,crossing_f_B,crossing_r_B=get_region(samfile,chrB,sv_clusters[chrA][chrB][cluster]["startB"],sv_clusters[chrA][chrB][cluster]["endB"],posB,min_mapq,max_ins_len,contig_number)

				sample_data[sample]={}
				sample_data[sample]={"covA":coverageA,"QA":frac_low_qA,"discA":n_discsA,"splitA":n_splitsA,"refRA":crossing_r_A,"refFA":crossing_f_A}
				sample_data[sample].update({"covB":coverageB,"QB":frac_low_qB,"discB":n_discsB,"splitB":n_splitsB,"refRB":crossing_r_B,"refFB":crossing_f_B})

				if chrA != chrB:
					sample_data[sample]["covM"]=0
				elif abs(posB - posA) < 1000:
					if posA < posB:
						coverageM,_,_,_,_,_=get_region(samfile,chrA,posA,posB,posA,min_mapq,max_ins_len,contig_number)
					else:
						coverageM,_,_,_,_,_=get_region(samfile,chrA,posB,posA,posB,min_mapq,max_ins_len,contig_number)

					sample_data[sample]["covM"]=coverageM
				else:
					s=int(math.floor(posA/50.0))
					e=int(math.floor(posB/50.0))+1
					sample_data[sample]["covM"]=numpy.average(coverage_data[chrA][s:e] )

			inverted=0
			non_inverted=0
			for i in range(0,len(sv_clusters[chrA][chrB][cluster]["positions_A"]["orientation_discordants"]) ):
				if sv_clusters[chrA][chrB][cluster]["positions_A"]["orientation_discordants"][i] == sv_clusters[chrA][chrB][cluster]["positions_B"]["orientation_discordants"][i]:
					inverted+=1
				else:
					non_inverted+=1

			for i in range(0,len(sv_clusters[chrA][chrB][cluster]["positions_A"]["orientation_splits"]) ):
				if not sv_clusters[chrA][chrB][cluster]["positions_A"]["orientation_splits"][i] == sv_clusters[chrA][chrB][cluster]["positions_B"]["orientation_splits"][i]:
					inverted+=1
				else:
					non_inverted+=1

			for i in range(0,len(sv_clusters[chrA][chrB][cluster]["positions_A"]["orientation_contigs"]) ):
				if not sv_clusters[chrA][chrB][cluster]["positions_A"]["orientation_contigs"][i] == sv_clusters[chrA][chrB][cluster]["positions_B"]["orientation_contigs"][i]:
					inverted+=1
				else:
					non_inverted+=1

			svtype,cn=find_sv_type(chrA,chrB,inverted,non_inverted,args,sample_data,samples,library)

			filt=sv_filter(sample_data,args,chrA,chrB,posA,posB,max_ins_len,n_discordants,n_splits,library,sample_data[sample]["discA"],sample_data[sample]["discB"],sample_data[sample]["splitA"],sample_data[sample]["splitB"],n_contigs)
			format_col="GT:CN:COV:DV:RV:LQ:RR:DR"

			#configure filters for CNV based on Read depth
			for sample in samples:

				covA=sample_data[sample]["covA"]
				covM=sample_data[sample]["covM"]
				covB=sample_data[sample]["covB"]

				if "DEL" in svtype:
					#homozygout del based on coverage
					if cn == 0:
						filt="PASS"


					#normal coverage on the flanking regions, abnormal inbetween
					if covA > covM*(cn+0.9) and covB > covM*(cn+0.9):
						filt="PASS"

				#too few reads, but clear DR signal
				elif "DUP" in svtype and filt == "BelowExpectedLinks":
					filt="PASS"
				scoring_dict={"n_contigs":n_contigs, "n_discordants":n_discordants,"n_splits":n_splits,"covA":covA,"covM":covM,"covB":covB,"refRA":sample_data[sample]["refRA"],"refRB":sample_data[sample]["refRB"],"refFA":sample_data[sample]["refFA"],"refFB":sample_data[sample]["refFB"]}

			if svtype != "BND":
				info=["SVTYPE={}".format(svtype),"SVLEN={}".format(posB-posA),"END={}".format(posB)]
				alt="<{}>".format(svtype)

				info+=["REGIONA={},{}".format(sv_clusters[chrA][chrB][cluster]["startA"],sv_clusters[chrA][chrB][cluster]["endA"])]
				info+=["REGIONB={},{}".format(sv_clusters[chrA][chrB][cluster]["startB"],sv_clusters[chrA][chrB][cluster]["endB"])]
				info+=["LFA={},{}".format(sample_data[sample]["discA"],sample_data[sample]["splitA"])]
				info+=["LFB={},{}".format(sample_data[sample]["discB"],sample_data[sample]["splitB"])]
				info+=["LTE={},{}".format(n_discordants,n_splits)]

				if n_contigs:
					for c in sv_clusters[chrA][chrB][cluster]["contigs"]:
						if "_" in c:
							c=c.split("_")[0]
						ctgs=[ contig_seqs[c] ]
					info+=["CTG={}".format("|".join(ctgs) )]

				else:
					info+=["CTG=."]



				info=";".join(info)
				variant=[chrA,str(posA),"SV_{}_1".format(var_n),"N",alt,".",filt,info,format_col]
				for sample in samples:
					GT="./."

					if len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample]) >= args.r or len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]) >= args.p:
						GT="0/1"
					if sample_data[sample]["refRB"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample]) or sample_data[sample]["refRA"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample]):
						GT="1/1"
					if sample_data[sample]["refFB"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]) or sample_data[sample]["refFA"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]):
						GT="1/1"
					if n_contigs and (not len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]) and not len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample])):
						if sample_data[sample]["covB"]:
							if sample_data[sample]["refRB"]/sample_data[sample]["covB"] < 0.2:
								GT="1/1"
							else:
								GT="0/1"
						else:
							GT="1/1"

						if sample_data[sample]["covA"]:
							if sample_data[sample]["refRA"]/sample_data[sample]["covA"] < 0.2:
								GT="1/1"
							else:
								GT="0/1"
						else:
							GT="1/1"


					if "DEL" in alt:
						if cn == 0:
							GT = "1/1"
						else:
							GT= "0/1"
					elif "DUP" in alt:
						if cn >= 2*library["contig_ploidy_{}".format(chrA)]:
							GT= "1/1"
						else:
							GT="0/1"

					variant.append( "{}:{}:{},{},{}:{}:{}:{},{}:{},{}:{},{}".format(GT,cn,sample_data[sample]["covA"],sample_data[sample]["covM"],sample_data[sample]["covB"],n_discordants,n_splits,sample_data[sample]["QA"],sample_data[sample]["QB"],sample_data[sample]["refRA"],sample_data[sample]["refRB"],sample_data[sample]["refFA"],sample_data[sample]["refFB"]) )
				variants.append([chrA,posA,variant,scoring_dict])
			else:
				info=["SVTYPE=BND".format(svtype)]
				inverted=False
				before=True

				if posA == sv_clusters[chrA][chrB][cluster]["endA"]:
					before=False

				if inverted  > non_inverted:
					inverted=True

				if not inverted and not before:
					alt_str_a="N[{}:{}[".format(chrB,posB)
					alt_str_b="]{}:{}]N".format(chrA,posA)
				elif not inverted and before:
					alt_str_a="]{}:{}]N".format(chrB,posB)
					alt_str_b="N[{}:{}[".format(chrA,posA)
				elif inverted and  not before:
					alt_str_a="N]{}:{}]".format(chrB,posB)
					alt_str_b="[{}:{}[N".format(chrA,posA)
				else:
					alt_str_a="[{}:{}[N".format(chrB,posB)
					alt_str_b="N]{}:{}]".format(chrA,posA)

				info+=["REGIONA={},{}".format(sv_clusters[chrA][chrB][cluster]["startA"],sv_clusters[chrA][chrB][cluster]["endA"])]
				info+=["REGIONB={},{}".format(sv_clusters[chrA][chrB][cluster]["startB"],sv_clusters[chrA][chrB][cluster]["endB"])]
				info+=["LFA={},{}".format(sample_data[sample]["discA"],sample_data[sample]["splitA"])]
				info+=["LFB={},{}".format(sample_data[sample]["discA"],sample_data[sample]["splitA"])]
				info+=["LTE={},{}".format(n_discordants,n_splits)]

				if n_contigs:
					for c in sv_clusters[chrA][chrB][cluster]["contigs"]:
						if "_" in c:
							c=c.split("_")[0]

						ctgs=[ contig_seqs[c] ]
					info+=["CTG={}".format("|".join(ctgs) )]

				else:
					info+=["CTG=."]



				info=";".join(info)
				variant=[chrA,str(posA),"SV_{}_1".format(var_n),"N",alt_str_a,".",filt,info,format_col]
				for sample in samples:
					GT="./."
					if len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample]) >= args.r or len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]) >= args.p:
						GT="0/1"
					if sample_data[sample]["refRB"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample]) or sample_data[sample]["refRA"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample]):
						GT="1/1"
					if sample_data[sample]["refFB"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]) or sample_data[sample]["refFA"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]):
						GT="1/1"
					if n_contigs and (not len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]) and not len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample])):
						if sample_data[sample]["covB"]:
							if sample_data[sample]["refRB"]/sample_data[sample]["covB"] < 0.2:
								GT="1/1"
							else:
								GT="0/1"
						else:
								GT="1/1"

						if sample_data[sample]["covA"]:
							if sample_data[sample]["refRA"]/sample_data[sample]["covA"] < 0.2:
								GT="1/1"
							else:
								GT="0/1"


						else:
								GT="1/1"



					variant.append( "{}:{}:{},{},{}:{}:{}:{},{}:{},{}:{},{}".format(GT,cn,sample_data[sample]["covA"],sample_data[sample]["covM"],sample_data[sample]["covB"],n_discordants,n_splits,sample_data[sample]["QA"],sample_data[sample]["QB"],sample_data[sample]["refRA"],sample_data[sample]["refRB"],sample_data[sample]["refFA"],sample_data[sample]["refFB"]) )
				variants.append([chrA,posA,variant,scoring_dict])


				variant=[chrB,str(posB),"SV_{}_2".format(var_n),"N",alt_str_b,".",filt,info,format_col]
				for sample in samples:
					GT="./."
					if len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample]) >= args.r or len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]) >= args.p:
						GT="0/1"
					if sample_data[sample]["refRB"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample]) or sample_data[sample]["refRA"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample]):
						GT="1/1"
					if sample_data[sample]["refFB"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]) or sample_data[sample]["refFA"] < 0.1*len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]):
						GT="1/1"
					if n_contigs and (not len(sv_clusters[chrA][chrB][cluster]["sample_discordants"][sample]) and not len(sv_clusters[chrA][chrB][cluster]["sample_splits"][sample])):
						if sample_data[sample]["covB"]:
							if sample_data[sample]["refRB"]/sample_data[sample]["covB"] < 0.2:
								GT="1/1"
							else:
								GT="0/1"
						else:
								GT="1/1"

						if sample_data[sample]["covA"]:
							if sample_data[sample]["refRA"]/sample_data[sample]["covA"] < 0.2:
								GT="1/1"
							else:
								GT="0/1"
						else:
								GT="1/1"




					variant.append( "{}:{}:{},{},{}:{}:{}:{},{}:{},{}:{},{}".format(GT,cn,sample_data[sample]["covA"],sample_data[sample]["covM"],sample_data[sample]["covB"],n_discordants,n_splits,sample_data[sample]["QA"],sample_data[sample]["QB"],sample_data[sample]["refRA"],sample_data[sample]["refRB"],sample_data[sample]["refFA"],sample_data[sample]["refFB"]) )
				variants.append([chrB,posB,variant, scoring_dict ])

	samfile.close()
	return(variants)

def main(str bam_file_name,dict sv_clusters,args,dict library,int min_mapq,samples,dict coverage_data,contig_number,max_ins_len):
	contig_seqs={}
	new_seq=False
	if not args.skip_assembly:
		for line in open("{}_tiddit/clips.fa.assembly.clean.mag".format(args.o)):

			if not new_seq and line[0] == "@" and "\t" in line:
				name=line.split("\t")[0][1:]
				new_seq=True

			elif new_seq:
				contig_seqs[name]=line.strip("\n")
				new_seq=False
	variants={}
	for chrA in sv_clusters:
		variants[chrA]=[]
		for chrB in sv_clusters[chrA]:
			variants[chrB]=[]

	variants_list=Parallel(n_jobs=args.threads)( delayed(define_variant)(chrA,bam_file_name,sv_clusters,args,library,min_mapq,samples,coverage_data,contig_number,max_ins_len,contig_seqs) for chrA in sv_clusters)

	ratios={"fragments_A":[],"fragments_B":[],"reads_A":[],"reads_B":[]}
	for v in variants_list:
		for variant in v:
			if variant[3]["n_discordants"]:
				ratios["fragments_A"].append(variant[3]["n_discordants"]/(variant[3]["refFA"]+variant[3]["n_discordants"]) )
				ratios["fragments_B"].append(variant[3]["n_discordants"]/(variant[3]["refFB"]+variant[3]["n_discordants"]) )

			if variant[3]["n_splits"]:
				ratios["reads_A"].append(variant[3]["n_splits"]/(variant[3]["refRA"]+variant[3]["n_splits"]) )
				ratios["reads_B"].append(variant[3]["n_splits"]/(variant[3]["refRB"]+variant[3]["n_splits"]) )


	p=[1,5,10,20,30,40,50,60,70,75,80,85,90,95,97.5,99]

	percentiles={"FA":percentile(ratios["fragments_A"],p),"FB":percentile(ratios["fragments_B"],p),"RA":percentile(ratios["reads_A"],p),"RB":percentile(ratios["reads_B"],p)}

	for v in variants_list:
		for variant in v:	
			score=scoring(variant[3],percentiles)
			variant[2][5]=str(score)
			variants[ variant[0] ].append( [ variant[1],variant[2] ] )

	return(variants)



import numpy
import re
cimport numpy
import pysam
import gzip

import tiddit.tiddit_coverage as tiddit_coverage


def get_gc(str reference_fasta,bam_header,bin_size):

	gc=tiddit_coverage.create_coverage(bam_header,bin_size)[0]

	reference={}
	chromosomes=[]
	if not reference_fasta.endswith(".gz"):
		with open(reference_fasta, 'r') as f:
			sequence=f.read()
	else:
		with gzip.open(reference_fasta, 'r') as f:
			sequence=f.read()


	#split_reference=sequence.split(">")
	split_reference=re.split("\n>|^>", sequence)
	del sequence
	del split_reference[0]

	for chromosome in split_reference:
		content=chromosome.split("\n",1)
		reference[content[0].strip().split()[0]]=content[1].replace("\n","")

	N=set(["N","n"])
	GC=set(["G","g","C","c"])

	for chromosome in gc:
		for i in range(0,len(gc)):
			start=i*bin_size
			end=(i+1)*bin_size
			if end > len(reference[chromosome]):
				end=len(reference[chromosome])

			N_count=0
			GC_count=0
			for j in range(start,end):
				if reference[chromosome][j] in N:
					N_count +=1

			#masked bin
			if N_count > bin_size/2:
				gc[chromosome][i] = -1	

	return(gc)


def determine_ploidy(dict coverage_data,contigs,dict library,int ploidy,str prefix,c, str reference_fasta,int bin_size,bam_header):

	gc=get_gc(reference_fasta,bam_header,bin_size)

	f=open( "{}.ploidies.tab".format(prefix),"w" )
	f.write("Chromosome\tPloidy\tPloidy_rounded\tMean_coverage\n")
	all_cov=[]
	for chromosome in coverage_data:
		tmp=[]
		for i in range(0,len(coverage_data[chromosome])):
			if coverage_data[chromosome][i] > 0 and gc[chromosome][i] != -1:

				tmp.append(coverage_data[chromosome][i])
				all_cov.append(coverage_data[chromosome][i])
		
		library[ "avg_coverage_{}".format(chromosome) ]=numpy.median(tmp)
		if numpy.isnan(library[ "avg_coverage_{}".format(chromosome) ]):
			library[ "avg_coverage_{}".format(chromosome) ]=0

	if not c:
		library["avg_coverage"]=numpy.median(all_cov)
	else:
		library["avg_coverage"]=c

	for chromosome in contigs:
		if not chromosome in coverage_data:
			continue

		avg_coverage_contig=library[ "avg_coverage_{}".format(chromosome) ]
		library["contig_ploidy_{}".format(chromosome)]=int(round(ploidy*avg_coverage_contig/library["avg_coverage"]))
		f.write("{}\t{}\t{}\t{}\n".format(chromosome,avg_coverage_contig/library["avg_coverage"]*ploidy,library["contig_ploidy_{}".format(chromosome)],avg_coverage_contig))
	

	f.close()
	return(library)




import sys
import time
cimport numpy
import numpy
import math
cimport cython
@cython.boundscheck(False)
@cython.wraparound(False)

def create_coverage(bam_header,bin_size,c="all"):
	coverage_data={}
	end_bin_size={}

	for contig in bam_header["SQ"]:
		if c == "all" or contig["SN"] == c:
			bins= int(math.ceil(contig["LN"]/float(bin_size)))
			coverage_data[ contig["SN"] ]=numpy.zeros(bins)
			end_bin_size[contig["SN"]]=contig["LN"]-(bins-1)*bin_size
			if c != "all":
				return(coverage_data[ contig["SN"] ],end_bin_size[contig["SN"]])
	return(coverage_data,end_bin_size)
def print_coverage(coverage_data,bam_header,bin_size,file_type,outfile):
	f=open(outfile,"w",buffering=819200)

	if file_type == "bed":
		f.write("#chromosome\tstart\tend\tcoverage\n")
	elif file_type == "wig":
		f.write("track type=wiggle_0 name=\"Coverage\" description=\"Per bin average coverage\"\n")

	for contig in bam_header["SQ"]:
		if file_type == "wig":
			f.write("fixedStep chrom={} start=1 step={}\n".format(contig["SN"],bin_size))

		for i in range(0,len(coverage_data[ contig["SN"] ]) ) :
			if file_type == "bed":
				bin_end=(i+1)*bin_size+1

				if i == len(coverage_data[ contig["SN"] ]) -1:
					bin_end=contig["LN"]

				f.write("{}\t{}\t{}\t{}\n".format(contig["SN"],1+i*bin_size,bin_end,coverage_data[ contig["SN"] ][i] ) )
			elif file_type == "wig":
				f.write("{}\n".format( coverage_data[ contig["SN"] ][i] ))

	f.close()

ctypedef numpy.double_t DTYPE_t
def update_coverage(long ref_start,long ref_end,int bin_size,numpy.ndarray[DTYPE_t, ndim=1] coverage_data,int end_bin_size):

	cdef int first_bin=ref_start//bin_size
	cdef int end_bin=(ref_end-1)//bin_size

	cdef float bases_first_bin

	if end_bin == first_bin:
		bases_first_bin=ref_end-ref_start
		coverage_data[first_bin]=bases_first_bin/bin_size+coverage_data[first_bin]

		return(coverage_data)

	bases_first_bin=((first_bin+1)*bin_size)-ref_start
	coverage_data[first_bin]=bases_first_bin/bin_size+coverage_data[first_bin]
	cdef float bases_last_bin=(ref_end-1)-end_bin*bin_size


	if end_bin < len(coverage_data)-1:
		coverage_data[end_bin]=bases_last_bin/bin_size+coverage_data[end_bin]
	else:
		coverage_data[end_bin]=bases_last_bin/end_bin_size+coverage_data[end_bin]

	for i in range(first_bin+1,end_bin):
		coverage_data[i]=1.0+coverage_data[i]

	return(coverage_data)

#bam_file_name=sys.argv[1]

#samfile = pysam.AlignmentFile(bam_file_name, "r")
#bam_header=samfile.header
#bin_size=50
#file_type="bed"
#outfile=sys.argv[2]
#min_q=10

#coverage_data,end_bin_size=create_coverage(bam_header,bin_size)

#for read in samfile.fetch():
#	coverage_data=update_coverage(read,bin_size,coverage_data,min_q,end_bin_size)

#print_coverage(coverage_data,bam_header,bin_size,file_type,outfile)

#samfile.close()



import pysam
import numpy
def statistics(bam_file_name,ref,min_mapq,max_ins_len,n_reads):
	library={}
	samfile = pysam.AlignmentFile(bam_file_name, "r",reference_filename=ref)

	insert_size=[]
	read_length=[]
	is_innie=0
	is_outtie=0

	n_sampled=0
	for read in samfile.fetch():
	
		if read.mate_is_unmapped:
			continue

		if read.is_reverse == read.mate_is_reverse:
			continue

		if read.next_reference_name != read.reference_name or (read.template_length > max_ins_len):
			continue

		if read.next_reference_start < read.reference_start:
			continue	

		if read.is_supplementary or read.is_secondary or read.is_duplicate or read.mapq < min_mapq:
			continue

		n_sampled+=1

		insert_size.append( read.template_length )
		read_length.append( read.query_length )

		if read.is_reverse and not read.mate_is_reverse:
			is_outtie+=1
		else:
			is_innie+=1

		if n_sampled > n_reads:
			break

	samfile.close()

	library["avg_read_length"]=numpy.average(read_length)
	library["avg_insert_size"]=numpy.average(insert_size)
	library["std_insert_size"]=numpy.std(insert_size)
	library["percentile_insert_size"]=numpy.percentile(insert_size, 99.9)

	print("LIBRARY STATISTICS")
	if is_innie > is_outtie:
		library["mp"]=False
		print("\tPair orientation = Forward-Reverse")
	else:
		print("\tPair orientation = Reverse-Forward")
		library["mp"]=True


	print("\tAverage Read length = {}".format(library["avg_read_length"]) )
	print("\tAverage insert size = {}".format(library["avg_insert_size"]) )
	print("\tStdev insert size = {}".format(library["std_insert_size"] ) )
	print("\t99.95 percentile insert size = {}".format( library["percentile_insert_size"]) )
	print("")

	return(library)



import pysam
import sys
import os
import itertools
import time
from joblib import Parallel, delayed

import tiddit.tiddit_coverage as tiddit_coverage

def find_SA_query_range(SA):
	a =pysam.AlignedSegment()
	a.reference_start=int( SA[1] )

	if SA[2] == "+":
		a.flag = 64
	else:
		a.flag = 80

	cdef list SA_cigar=[]
	SC = ["".join(x) for _, x in itertools.groupby(SA[3], key=str.isdigit)]		

	cdef dict s_to_op={"M":0,"S":4,"H":5,"D":2,"I":1}
	for i in range(0,int(len(SC)/2)):
		op=s_to_op[SC[i*2+1]]
		SA_cigar.append( (op,int(SC[i*2])) )

	a.cigar = tuple(SA_cigar)
	return(a)

def SA_analysis(read,min_q,tag,reference_name):
	#print(read.query_alignment_start,read.query_alignment_end,read.is_reverse,read.cigarstring)
	suplementary_alignments=read.get_tag(tag).rstrip(";").split(";")
	
	if len(suplementary_alignments) > 1:
		SA_lengths=[]
		ok_q=[]
		for i in range(0,len(suplementary_alignments)):
			SA_data=suplementary_alignments[0].split(",")
			if int(SA_data[4]) >= min_q:
				ok_q.append(i)
				supplementry_alignment=find_SA_query_range(SA_data)
				SA_lengths.append(supplementry_alignment.query_alignment_end-supplementry_alignment.query_alignment_start)

		longest_aln=0
		for i in range(0,len(ok_q)):
			if SA_lengths[i] > SA_lengths[longest_aln]:
				longest_aln=i

		#all alignments fail quality treshold
		if len(ok_q) == 0:
			return()

		#only one SA pass mapping quality treshold		
		elif len(ok_q) == 1:
			suplementary_alignments[0]=suplementary_alignments[ ok_q[0] ]
		#many SA pass mapping quality treshold, pick the longest alignment.
		else:
			suplementary_alignments[0]=suplementary_alignments[ longest_aln ]

	SA_data=suplementary_alignments[0].split(",")
	SA_pos=int(SA_data[1])

	if int(SA_data[4]) < min_q:
		return()

	cdef long read_query_alignment_end=read.query_alignment_end

	clip_before=False

	supplementry_alignment=find_SA_query_range(SA_data)
	if supplementry_alignment.query_alignment_start < read.query_alignment_start:
		clip_before=True

	if not clip_before:
		if read.is_reverse:


			split_pos=read.reference_start+1
		else:
			split_pos=read.reference_end+1		
	else:
		if read.is_reverse:
			split_pos=read.reference_end+1
		else:
			split_pos=read.reference_start+1

	SA_chr=SA_data[0]

	startA=read.reference_start+1
	endA=read.reference_end+1

	startB=supplementry_alignment.reference_start
	endB=supplementry_alignment.reference_end

	if clip_before:
		if SA_data[2] == "-":

			SA_split_pos=supplementry_alignment.reference_start
		else:
			SA_split_pos=supplementry_alignment.reference_end
	else:
		if SA_data[2] == "-":
			SA_split_pos=supplementry_alignment.reference_end

		else:
			SA_split_pos=supplementry_alignment.reference_start


	if SA_chr < reference_name:
		chrA=SA_chr
		chrB=reference_name
		tmp=split_pos
		split_pos=SA_split_pos
		SA_split_pos=tmp

		startB=read.reference_start+1
		endB=read.reference_end+1
		startA=supplementry_alignment.reference_start
		endA=supplementry_alignment.reference_end


	else:
		chrA=reference_name
		chrB=SA_chr

		if chrA == chrB:
			if SA_split_pos < split_pos:
				tmp=split_pos
				split_pos=SA_split_pos
				SA_split_pos=tmp

				startB=read.reference_start+1
				endB=read.reference_end+1
				startA=supplementry_alignment.reference_start
				endA=supplementry_alignment.reference_end

	split=[]
	if "-" == SA_data[2]:
		split=[chrA,chrB,read.query_name,split_pos,read.is_reverse,SA_split_pos, True,startA,endA,startB,endB]
	else:
		split=[chrA,chrB,read.query_name,split_pos,read.is_reverse,SA_split_pos,False,startA,endA,startB,endB]
		#splits[chrA][chrB][read.query_name]+=[split_pos,read.is_reverse,SA_split_pos,False]

	return(split)

def worker(str chromosome, str bam_file_name,str ref,str prefix,int min_q,int max_ins,str sample_id, int bin_size):
	print("Collecting signals on contig: {}".format(chromosome))
	samfile = pysam.AlignmentFile(bam_file_name, "r",reference_filename=ref,index_filename="{}_tiddit/{}.csi".format(prefix,sample_id))
	bam_header=samfile.header
	coverage_data,end_bin_size=tiddit_coverage.create_coverage(bam_header,bin_size,chromosome)	

	clips=[]
	data=[]
	splits=[]

	clip_dist=100

	cdef long read_position
	cdef long read_end
	cdef int mapq

	for read in samfile.fetch(chromosome,until_eof=True):

		if read.is_unmapped or read.is_duplicate:
			continue

		read_chromosome=read.reference_name
		mate_chromosome=read.next_reference_name
		read_position=read.reference_start
		read_end=read.reference_end
		read_mapq=read.mapq
		read_supplementary=read.is_supplementary

		if read_mapq >= min_q:
			coverage_data=tiddit_coverage.update_coverage(read_position,read_end,bin_size,coverage_data,end_bin_size)

		if read_supplementary or read.is_secondary:
			continue

		if read_mapq > 1:
			cigar_tuple=read.cigartuples
			if (cigar_tuple[0][0] == 4 and cigar_tuple[0][1] > 10) and (cigar_tuple[-1][0] == 0 and cigar_tuple[-1][1] > 30) and len(cigar_tuple) < 7:
				clips.append([">{}|{}|{}\n".format(read.query_name,read_chromosome,read_position+1),read.query_sequence+"\n"])

			elif cigar_tuple[-1][0] == 4 and cigar_tuple[-1][1] > 10 and (cigar_tuple[0][0] == 0 and cigar_tuple[0][1] > 30) and len(cigar_tuple) < 7:
				clips.append([">{}|{}|{}\n".format(read.query_name,read_chromosome,read_position+1),read.query_sequence+"\n"])

		if read_mapq < min_q:
			continue

		if read.has_tag("SA"):
			split=SA_analysis(read,min_q,"SA",read_chromosome)
			if split:
				splits.append(split)

		if read.mate_is_unmapped:
			continue

		if not read.is_paired:
			continue


		if ( abs(read.isize) > max_ins or mate_chromosome != read_chromosome ):
			read_query_name=read.query_name

			if mate_chromosome < read_chromosome:
				chrA=mate_chromosome
				chrB=read_chromosome
			else:
				chrA=read_chromosome
				chrB=mate_chromosome

			data.append([chrA,chrB,read_query_name,read_position+1,read_end+1,read.is_reverse,read_chromosome])

	f=open("{}_tiddit/clips/{}.fa".format(prefix,chromosome),"w")
	for clip in clips:
		f.write("".join(clip))
	f.close()

	return(chromosome,data,splits,coverage_data, "{}_tiddit/clips/{}.fa".format(prefix,chromosome) )

def main(str bam_file_name,str ref,str prefix,int min_q,int max_ins,str sample_id, int threads, int min_contig):

	samfile = pysam.AlignmentFile(bam_file_name, "r",reference_filename=ref,index_filename="{}_tiddit/{}.csi".format(prefix,sample_id))
	bam_header=samfile.header
	samfile.close()
	cdef int bin_size=50
	cdef str file_type="wig"
	cdef str outfile=prefix+".tiddit_coverage.wig"

	t_tot=0

	cdef dict data={}
	cdef dict splits={}
	cdef dict coverage_data={}
	cdef list clip_fasta=[]
	chromosomes=[]
	
	for chrA in bam_header["SQ"]:
		if chrA["LN"] < min_contig:
			continue

		chromosomes.append(chrA["SN"])
		data[chrA["SN"]]={}
		splits[chrA["SN"]]={}
		for chrB in bam_header["SQ"]:
			data[chrA["SN"]][chrB["SN"]]={}
			splits[chrA["SN"]][chrB["SN"]]={}

	t=time.time()
	res=Parallel(n_jobs=threads)( delayed(worker)(chromosome,bam_file_name,ref,prefix,min_q,max_ins,sample_id,bin_size) for chromosome in chromosomes )

	chromosomes=set(chromosomes)
	for i in range(0,len(res)):
		coverage_data[ res[i][0] ] = res[i][3]

		if not res[i][0] in chromosomes:
			continue

		for signal in res[i][1]:
			if not signal[0] in data:
				continue
	
			if not signal[2] in data[ signal[0] ][ signal[1] ]:
				data[ signal[0] ][signal[1]][signal[2]]=[]
			data[ signal[0] ][signal[1]][signal[2]].append(signal[3:])

		for signal in res[i][2]:
			if not signal[0] in splits:
				continue

			if not signal[2] in splits[ signal[0] ][ signal[1] ]:
				splits[ signal[0] ][signal[1]][signal[2]]=[]
			splits[ signal[0] ][signal[1]][signal[2]]+=signal[3:]

		clip_fasta.append(res[i][4])

	t_tot=time.time()-t

	print("total",t_tot)
	#print("coverage",t_update)
	#print("split",t_split)
	#print("disc",t_disc)

	#print("writing coverage wig")
	#tiddit_coverage.print_coverage(coverage_data,bam_header,bin_size,file_type,outfile)

	print("Writing signals to file")

	f=open("{}_tiddit/discordants_{}.tab".format(prefix,sample_id),"w")

	for chrA in data:
		for chrB in data[chrA]:
			for fragment in data[chrA][chrB]:
				if len(data[chrA][chrB][fragment]) < 2:
					continue

				if chrA == chrB:
					if data[chrA][chrB][fragment][1][-1] < data[chrA][chrB][fragment][0][-1]:
						out=data[chrA][chrB][fragment][1][0:-1]+data[chrA][chrB][fragment][0][0:-1]
					else:
						out=data[chrA][chrB][fragment][0][0:-1]+data[chrA][chrB][fragment][1][0:-1]
				else:
					if data[chrA][chrB][fragment][0][-1] == chrA:
						out=data[chrA][chrB][fragment][0][0:-1]+data[chrA][chrB][fragment][1][0:-1]
					else:
						out=data[chrA][chrB][fragment][1][0:-1]+data[chrA][chrB][fragment][0][0:-1]

				f.write("{}\t{}\t{}\t{}\n".format(fragment,chrA,chrB,"\t".join(map(str, out )))  )
	f.close()

	f=open("{}_tiddit/splits_{}.tab".format(prefix,sample_id),"w")

	for chrA in splits:
		for chrB in splits[chrA]:
			for fragment in splits[chrA][chrB]:
				f.write("{}\t{}\t{}\t{}\n".format(fragment,chrA,chrB,"\t".join(map(str, splits[chrA][chrB][fragment] )))  )
	f.close()

	f=open("{}_tiddit/clips_{}.fa".format(prefix,sample_id),"w")
	for clips in clip_fasta:
		for clip in open(clips):
			f.write(clip)
	f.close()

	return(coverage_data)



import pysam
import numpy
import math
import os

import tiddit.DBSCAN as DBSCAN
import tiddit.tiddit_signal as tiddit_signal


def read_contigs(aligned_contigs,prefix,sample_id,min_size):
	samfile = pysam.AlignmentFile(aligned_contigs, "r")
	bam_header=samfile.header
	

	split_contigs={}
	for chrA in bam_header["SQ"]:
		for chrB in bam_header["SQ"]:
			if chrA["SN"] <= chrB["SN"]:

				if not chrA["SN"] in split_contigs:
					split_contigs[chrA["SN"]] = {}

				split_contigs[chrA["SN"]][chrB["SN"]]={}

	for read in samfile.fetch(until_eof=True):

		if read.is_unmapped:
			continue

		if read.has_tag("SA") and not (read.is_supplementary or read.is_secondary):
			split=tiddit_signal.SA_analysis(read,-2,"SA",read.reference_name)

			if split:
				if not split[2] in split_contigs[split[0]][split[1]]:
					split_contigs[split[0]][split[1]][split[2]]=[]
				split_contigs[split[0]][split[1]][split[2]]+=split[3:]

			
		elif read.has_tag("XA") and not (read.is_supplementary or read.is_secondary):
			XA=read.get_tag("XA")
			if XA.count(";") == 1:
				if ",-" in XA:
					XA=XA.replace(",-",",")
					xa_list=XA.split(",")
					xa_list.insert(2,"-")
					XA=",".join(xa_list)
				else:
					XA=XA.replace(",+",",")
					xa_list=XA.split(",")
					xa_list.insert(2,"+")
					XA=",".join(xa_list)

				read.set_tag("XA",XA)
				split=tiddit_signal.SA_analysis(read,-2,"XA",read.reference_name)

				if split:
					if not split[2] in split_contigs[split[0]][split[1]]:
						split_contigs[split[0]][split[1]][split[2]]=[]
					split_contigs[split[0]][split[1]][split[2]]+=split[3:]

		elif not (read.is_supplementary or read.is_secondary) and len(read.cigartuples) > 2:

			current_bp=read.reference_start
			for i in range(0,len(read.cigartuples)-1):
				if read.cigartuples[i][0] == 2 and read.cigartuples[i][1] > min_size:
					
					split_contigs[read.reference_name][read.reference_name]["{}_d_{}".format(read.query_name,i)]=[current_bp,read.is_reverse,current_bp+read.cigartuples[i][1],read.is_reverse,read.reference_start,current_bp,current_bp+read.cigartuples[i][1],read.reference_end]
				current_bp+=read.cigartuples[i][1]

	f=open("{}_tiddit/contigs_{}.tab".format(prefix,sample_id),"w")
	for chrA in split_contigs:
		for chrB in split_contigs[chrA]:
			for fragment in split_contigs[chrA][chrB]:

				f.write("{}\t{}\t{}\t{}\n".format(fragment,chrA,chrB,"\t".join(map(str, split_contigs[chrA][chrB][fragment] )))  )

	f.close()

def main(prefix,sample_id,library,contigs,coverage_data,args):
	clips={}
	c=[]

	for line in open("{}_tiddit/clips_{}.fa".format(prefix,sample_id)):

		if line[0] == ">":
			c.append(line.strip())
			pos=int(line.strip().split("|")[-1])
			chr=line.strip().split("|")[-2]
			if not chr in clips:
				clips[chr]=[[],[]]

		else:
			c.append(line.strip())
			clips[chr][0].append( "\n".join(c) )
			clips[chr][1].append([pos,0])
			c=[]

	f=open("{}_tiddit/clips.fa".format(prefix),"w")	
	for chr in clips:
		
		clusters,cluster_id = DBSCAN.x_coordinate_clustering(numpy.array(clips[chr][1]),50,args.l)
		cluster_stats={}

		for i in range(0,len(clusters)):
			if clusters[i] == -1:
				continue
			if not clusters[i] in cluster_stats:
				cluster_stats[clusters[i]]=[0,[]]
			cluster_stats[clusters[i]][0]+=1
			cluster_stats[clusters[i]][1].append( clips[chr][1][i][0] )

		for i in range(0,len(clusters)):
			if clusters[i] == -1:
				continue
			if cluster_stats[clusters[i]][0] < args.r:
				continue

			if cluster_stats[clusters[i]][0] > 2*library[ "avg_coverage_{}".format(chr) ]:
				continue

			clip_coverage=coverage_data[chr][ int(math.floor(clips[chr][1][i][0]/50.0)) ]
			if clip_coverage >  args.max_coverage/2*library[ "avg_coverage_{}".format(chr) ]:
				continue

			f.write( clips[chr][0][i].strip() +"\n")

	f.close()
	del clips

	os.system("{} -dNCr {}_tiddit/clips.fa | {} assemble -t {} -l 81 - > {}_tiddit/clips.fa.assembly.mag".format(args.ropebwt2,prefix,args.fermi2,args.threads,prefix))
	os.system("{} simplify -COS -d 0.8 {}_tiddit/clips.fa.assembly.mag 1> {}_tiddit/clips.fa.assembly.clean.mag 2> /dev/null".format(args.fermi2,prefix,prefix))
	os.system("{} mem -t {} -x intractg {} {}_tiddit/clips.fa.assembly.clean.mag  1> {}_tiddit/clips.sam 2> /dev/null".format(args.bwa,args.threads,args.ref,prefix,prefix))

	read_contigs("{}_tiddit/clips.sam".format(prefix) , prefix, sample_id, args.z)
	



import sys
import json
import argparse
import time
import pysam
import os
import shutil
import glob

import tiddit.tiddit_stats as tiddit_stats
import tiddit.tiddit_signal as tiddit_signal
import tiddit.tiddit_vcf_header as tiddit_vcf_header
import tiddit.tiddit_coverage_analysis as tiddit_coverage_analysis
import tiddit.tiddit_coverage as tiddit_coverage
import tiddit.tiddit_cluster as tiddit_cluster
import tiddit.tiddit_variant as tiddit_variant
import tiddit.tiddit_contig_analysis as tiddit_contig_analysis

def main():
	version="3.6.1"
	parser = argparse.ArgumentParser("""tiddit-{}""".format(version),add_help=False)
	parser.add_argument("--sv"	 , help="call structural variation", required=False, action="store_true")
	parser.add_argument("--cov"        , help="generate a coverage bed file", required=False, action="store_true")
	args, unknown = parser.parse_known_args()

	if args.sv == True:

		parser = argparse.ArgumentParser("""tiddit --sv --bam inputfile [-o prefix] --ref ref.fasta""")
		parser.add_argument('--sv'       , help="call structural variation", required=False, action="store_true")
		parser.add_argument('--bam', type=str,required=True, help="coordinate sorted bam file(required)")
		parser.add_argument('-o', type=str,default="output", help="output prefix(default=output)")
		parser.add_argument('-i', type=int, help="paired reads maximum allowed insert size. Pairs aligning on the same chr at a distance higher than this are considered candidates for SV (default= 99.9th percentile of insert size)")
		parser.add_argument('-d', type=str,help="expected reads orientations, possible values \"innie\" (-> <-) or \"outtie\" (<- ->). Default: major orientation within the dataset")
		parser.add_argument('-p', type=int,default=3, help="Minimum number of supporting pairs in order to call a variant (default 3)")
		parser.add_argument('--threads', type=int,default=1, help="Number of threads (default=1)")
		parser.add_argument('-r', type=int,default=3, help="Minimum number of supporting split reads to call a variant (default 3)")
		parser.add_argument('-q', type=int,default=5, help="Minimum mapping quality to consider an alignment (default 5)")
		parser.add_argument('-n', type=int,default=2, help="the ploidy of the organism,(default = 2)")
		parser.add_argument('-e', type=int, help="clustering distance  parameter, discordant pairs closer than this distance are considered to belong to the same variant(default = sqrt(insert-size*2)*12)")
		parser.add_argument('-c', type=float, help="average coverage, overwrites the estimated average coverage (useful for exome or panel data)")
		parser.add_argument('-l', type=int,default=3, help="min-pts parameter (default=3),must be set >= 2")
		parser.add_argument('-s', type=int,default=25000000, help="Number of reads to sample when computing library statistics(default=25000000)")
		parser.add_argument('-z', type=int,default=50, help="minimum variant size (default=50), variants smaller than this will not be printed ( z < 10 is not recomended)")
		parser.add_argument('--force_ploidy',action="store_true", help="force the ploidy to be set to -n across the entire genome (i.e skip coverage normalisation of chromosomes)")
		parser.add_argument('--n_mask',type=float,default=0.5, help="exclude regions from coverage calculation if they contain more than this fraction of N (default = 0.5)")
		parser.add_argument('--ref', type=str, help="reference fasta",required=True)
		parser.add_argument('--bwa', type=str,default="bwa", help="path to bwa executable file(default=bwa)")
		parser.add_argument('--fermi2', type=str,default="fermi2", help="path to fermi2 executable file (default=fermi2)")
		parser.add_argument('--ropebwt2', type=str , default="ropebwt2", help="path to ropebwt2 executable file (default=ropebwt2)")
		parser.add_argument('--skip_assembly', action="store_true", help="Skip running local assembly, tiddit will perform worse, but wont require fermi2, bwa, ropebwt and bwa indexed ref")
		parser.add_argument('--p_ratio', type=float,default=0.1, help="minimum discordant pair/normal pair ratio at the breakpoint junction(default=0.1)")
		parser.add_argument('--r_ratio', type=float,default=0.1, help="minimum split read/coverage ratio at the breakpoint junction(default=0.1)")
		parser.add_argument('--max_coverage', type=float,default=4, help="filter call if X times higher than chromosome average coverage (default=4)")
		parser.add_argument('--min_contig', type=int,default=10000, help="Skip calling on small contigs (default < 10000 bp)")
		args= parser.parse_args()

		if args.l < 2:
			print ("error, too low --l value!")
			quit()

		if not args.skip_assembly:
			if not os.path.isfile(args.bwa) and not shutil.which(args.bwa):
				print("error, BWA executable missing, add BWA to path, or specify using --bwa")
				quit()

			if not os.path.isfile(args.fermi2) and not shutil.which(args.fermi2):
				print("error, fermi2 executable missing, add fermi2 to path, or specify using --fermi2")
				quit()

			if not os.path.isfile(args.ropebwt2) and not shutil.which(args.ropebwt2):
				print("error, ropebwt2 executable missing, add ropebwt2 to path, or specify using --ropebwt2")
				quit()

			if not glob.glob("{}*.bwt*".format(args.ref)):
				print ("error, The reference must be indexed using bwa index; run bwa index, or skip local assembly (--skip_assembly)")
				quit()


		if not os.path.isfile(args.ref):
			print ("error,  could not find the reference file")
			quit()


		if not (args.bam.endswith(".bam") or args.bam.endswith(".cram")):
			print ("error, the input file is not a bam file, make sure that the file extension is .bam or .cram")
			quit()

		if not os.path.isfile(args.bam):
			print ("error,  could not find the bam file")
			quit()

		bam_file_name=args.bam
		samfile = pysam.AlignmentFile(bam_file_name, "r",reference_filename=args.ref)

		bam_header=samfile.header
		samfile.close()


		try:
			sample_id=bam_header["RG"][0]["SM"]
		except:
			sample_id=bam_file_name.split("/")[-1].split(".")[0]


		samples=[sample_id]

		contigs=[]
		contig_number={}
		contig_length={}
		i=0
		for contig in bam_header["SQ"]:
			contigs.append(contig["SN"])
			contig_number[contig["SN"]]=i
			contig_length[ contig["SN"] ]=contig["LN"]
			i+=1

		prefix=args.o
		os.mkdir(f"{prefix}_tiddit")
		os.mkdir(f"{prefix}_tiddit/clips")

		pysam.index("-c","-m","6","-@",str(args.threads),bam_file_name,"{}_tiddit/{}.csi".format(args.o,sample_id))

		min_mapq=args.q
		max_ins_len=100000
		n_reads=args.s

		library=tiddit_stats.statistics(bam_file_name,args.ref,min_mapq,max_ins_len,n_reads)
		if args.i:
			max_ins_len=args.i
		else:
			max_ins_len=library["percentile_insert_size"]


		t=time.time()
		coverage_data=tiddit_signal.main(bam_file_name,args.ref,prefix,min_mapq,max_ins_len,sample_id,args.threads,args.min_contig)
		print("extracted signals in:")
		print(t-time.time())

		t=time.time()
		library=tiddit_coverage_analysis.determine_ploidy(coverage_data,contigs,library,args.n,prefix,args.c,args.ref,50,bam_header)
		print("calculated coverage in:")
		print(time.time()-t)


		if not args.skip_assembly:

			t=time.time()
			tiddit_contig_analysis.main(prefix,sample_id,library,contigs,coverage_data,args)
			print("Clip read assembly in:")
			print(time.time()-t)

		vcf_header=tiddit_vcf_header.main( bam_header,library,sample_id,version )

		if not args.e:
			args.e=int(library["avg_insert_size"]/2.0)

		t=time.time()
		sv_clusters=tiddit_cluster.main(prefix,contigs,contig_length,samples,library["mp"],args.e,args.l,max_ins_len,args.min_contig,args.skip_assembly,args.r)

		print("generated clusters in")
		print(time.time()-t)

		f=open(prefix+".vcf","w")
		f.write(vcf_header+"\n")

		t=time.time()
		variants=tiddit_variant.main(bam_file_name,sv_clusters,args,library,min_mapq,samples,coverage_data,contig_number,max_ins_len)
		print("analyzed clusters in")
		print(time.time()-t)

		for chr in contigs:
			if not chr in variants:
				continue
			for variant in sorted(variants[chr], key=lambda x: x[0]):
				f.write( "\t".join(variant[1])+"\n" )
		f.close()
		quit()

	elif args.cov:
		parser = argparse.ArgumentParser("""tiddit --cov --bam inputfile [-o prefix]""")
		parser.add_argument('--cov'        , help="generate a coverage bed/wig file", required=False, action="store_true")
		parser.add_argument('--bam', type=str,required=True, help="coordinate sorted bam file(required)")
		parser.add_argument('-o', type=str,default="output", help="output prefix(default=output)")
		parser.add_argument('-z', type=int,default=500, help="use bins of specified size(default = 500bp) to measure the coverage of the entire bam file, set output to stdout to print to stdout")
		parser.add_argument('-w'        , help="generate wig instead of bed", required=False, action="store_true")
		parser.add_argument('-q'        , help="minimum mapping quality(default=20)", required=False, default=20)
		parser.add_argument('--ref', type=str, help="reference fasta, used for reading cram")
		args= parser.parse_args()

		if not os.path.isfile(args.bam):
			print ("error,  could not find the bam file")
			quit()

		samfile = pysam.AlignmentFile(args.bam, "r",reference_filename=args.ref)
		bam_header=samfile.header
		coverage_data,end_bin_size=tiddit_coverage.create_coverage(bam_header,args.z)
		n_reads=0
		for read in samfile.fetch(until_eof=True):

			if read.is_unmapped or read.is_duplicate:
				continue

			t=time.time()
			if read.mapq >= args.q:
				n_reads+=1

				read_position=read.reference_start
				read_end=read.reference_end
				read_reference_name=read.reference_name

				coverage_data[read_reference_name]=tiddit_coverage.update_coverage(read_position,read_end,args.z,coverage_data[read_reference_name],end_bin_size[read_reference_name])

		if args.w:
			tiddit_coverage.print_coverage(coverage_data,bam_header,args.z,"wig",args.o +".wig")
		else:
			tiddit_coverage.print_coverage(coverage_data,bam_header,args.z,"bed",args.o +".bed")

	else:
		parser.print_help()

if __name__ == '__main__':
	main()



import sys
import os
import tiddit.DBSCAN as DBSCAN
import numpy
import statistics 
from statistics import mode

def find_discordant_pos(fragment,is_mp):
	if is_mp:
		if fragment[5] == "False" and fragment[8] == "True":
			posA=fragment[3]
			posB=fragment[7]
		elif fragment[5] == "False" and fragment[8] == "False":
			posA=fragment[3]
			posB=fragment[6]
		elif fragment[5] == "True" and fragment[8] == "True":		
			posA=fragment[4]
			posB=fragment[7]
		else:
			posA=fragment[4]
			posB=fragment[6]

	else:
		if fragment[5] == "False" and fragment[8] == "True":
			posA=fragment[4]
			posB=fragment[6]
		elif fragment[5] == "False" and fragment[8] == "False":
			posA=fragment[4]
			posB=fragment[7]
		elif fragment[5] == "True" and fragment[8] == "True":
			posA=fragment[3]
			posB=fragment[6]

		else:
			posA=fragment[3]
			posB=fragment[7]

	return(posA,posB)

def main(prefix,chromosomes,contig_length,samples,is_mp,epsilon,m,max_ins_len,min_contig,skip_assembly,min_reads):

	discordants={}
	contigs=set([])
	splits={}
	positions={}

	i=0
	for sample in samples:
		for line in open("{}_tiddit/discordants_{}.tab".format(prefix,sample)):
			content=line.rstrip().split("\t")
			chrA=content[1]
			chrB=content[2]
			if contig_length[chrA] < min_contig or contig_length[chrB] < min_contig:
				continue

			if not chrA in positions:
				positions[chrA]={}
			if not chrB in positions[chrA]:
				positions[chrA][chrB]=[]

			if not chrA in discordants:
				discordants[chrA]={}
			if not chrB in discordants[chrA]:
				discordants[chrA][chrB]=[]

			posA,posB=find_discordant_pos(content,is_mp)

			if int(posA) > contig_length[chrA]:
				posA=contig_length[chrA]
				if int(posB) > contig_length[chrB]:
					posA=contig_length[chrB]
	
			discordants[chrA][chrB].append([content[0],sample,"D",posA,content[5],posB,content[8],i,int(content[3]),int(content[4]),int(content[6]),int(content[7])])
			positions[chrA][chrB].append([int(posA),int(posB),i])
			i+=1

		for line in open("{}_tiddit/splits_{}.tab".format(prefix,sample)):
			content=line.rstrip().split("\t")
			chrA=content[1]
			chrB=content[2]
			if contig_length[chrA] < min_contig or contig_length[chrB] < min_contig:
				continue

			if not chrA in positions:
				positions[chrA]={}
			if not chrB in positions[chrA]:
				positions[chrA][chrB]=[]

			if not chrA in discordants:
				discordants[chrA]={}
			if not chrB in discordants[chrA]:
				discordants[chrA][chrB]=[]

			posA=content[3]
			posB=content[5]

			if int(posA) > contig_length[chrA]:
				posA=contig_length[chrA]
			if int(posB) > contig_length[chrB]:
				posB=contig_length[chrB]

			discordants[chrA][chrB].append([content[0],sample,"S",posA,content[4],posB,content[6],i,int(content[7]),int(content[8]),int(content[9]),int(content[10])])
			positions[chrA][chrB].append([int(posA),int(posB),i])
			i+=1

		if not skip_assembly:
			for line in open("{}_tiddit/contigs_{}.tab".format(prefix,sample)):
				content=line.rstrip().split("\t")
				chrA=content[1]
				chrB=content[2]

				if contig_length[chrA] < min_contig or contig_length[chrB] < min_contig:
					continue


				if not chrA in positions:
					positions[chrA]={}
				if not chrB in positions[chrA]:
					positions[chrA][chrB]=[]

				if not chrA in discordants:
					discordants[chrA]={}
				if not chrB in discordants[chrA]:
					discordants[chrA][chrB]=[]


				posA=content[3]
				posB=content[5]

				if int(posA) > contig_length[chrA]:
					posA=contig_length[chrA]
				if int(posB) > contig_length[chrB]:
					posB=contig_length[chrB]

				discordants[chrA][chrB].append([content[0],sample,"A",posA,content[4],posB,content[6],i,int(content[7]),int(content[8]),int(content[9]),int(content[10])])
				positions[chrA][chrB].append([int(posA),int(posB),i])
				contigs.add(i)
				i+=1

	candidates={}
	for chrA in chromosomes:
		if not chrA in positions:
			continue
		if not chrA in candidates:
			candidates[chrA]={}

		for chrB in chromosomes:
			if not chrB in positions[chrA]:
				continue
			if not chrB in candidates[chrA]:
				candidates[chrA][chrB]={}

			positions[chrA][chrB]=numpy.array(sorted(positions[chrA][chrB],key=lambda l:l[0]))
			#print(positions[:,[0,1]])
			
			clusters=DBSCAN.main(positions[chrA][chrB],epsilon,m)
			positions[chrA][chrB]=list(positions[chrA][chrB])
			cluster_pos=[]
			for i in range(0,len(positions[chrA][chrB])):
				cluster_pos.append(list(positions[chrA][chrB][i])+[clusters[i]] )
				#positions[i].append(clusters[i])

			cluster_pos= sorted(cluster_pos,key=lambda l:l[2])
			n_ctg_clusters=0
			for i in range(0,len(cluster_pos)):
				candidate=int(cluster_pos[i][-1])
				if candidate == -1 and not (chrA == chrB and discordants[chrA][chrB][i][2] == "A" and ( int(discordants[chrA][chrB][i][5])- int(discordants[chrA][chrB][i][3]) ) < max_ins_len*2):
					continue
				elif candidate == -1 and discordants[chrA][chrB][i][2] == "A":
					candidate=len(cluster_pos)+n_ctg_clusters
					n_ctg_clusters+=1

				if not candidate in candidates[chrA][chrB]:
					candidates[chrA][chrB][candidate]={}
					candidates[chrA][chrB][candidate]["signal_type"]={}
					candidates[chrA][chrB][candidate]["samples"]=set([])
					candidates[chrA][chrB][candidate]["sample_discordants"]={}
					candidates[chrA][chrB][candidate]["sample_splits"]={}
					candidates[chrA][chrB][candidate]["sample_contigs"]={}


					candidates[chrA][chrB][candidate]["N_discordants"]=0
					candidates[chrA][chrB][candidate]["discordants"]=set([])
					candidates[chrA][chrB][candidate]["N_splits"]=0
					candidates[chrA][chrB][candidate]["splits"]=set([])
					candidates[chrA][chrB][candidate]["N_contigs"]=0
					candidates[chrA][chrB][candidate]["contigs"]=set([])


					candidates[chrA][chrB][candidate]["n_signals"]=0

					candidates[chrA][chrB][candidate]["posA"]=0
					candidates[chrA][chrB][candidate]["positions_A"]={}
					candidates[chrA][chrB][candidate]["positions_A"]["contigs"]=[]
					candidates[chrA][chrB][candidate]["positions_A"]["splits"]=[]
					candidates[chrA][chrB][candidate]["positions_A"]["discordants"]=[]
					candidates[chrA][chrB][candidate]["positions_A"]["orientation_contigs"]=[]
					candidates[chrA][chrB][candidate]["positions_A"]["orientation_splits"]=[]
					candidates[chrA][chrB][candidate]["positions_A"]["orientation_discordants"]=[]
					candidates[chrA][chrB][candidate]["positions_A"]["start"]=[]
					candidates[chrA][chrB][candidate]["positions_A"]["end"]=[]

					candidates[chrA][chrB][candidate]["start_A"]=0
					candidates[chrA][chrB][candidate]["end_A"]=0

					candidates[chrA][chrB][candidate]["posB"]=0
					candidates[chrA][chrB][candidate]["positions_B"]={}
					candidates[chrA][chrB][candidate]["positions_B"]["contigs"]=[]
					candidates[chrA][chrB][candidate]["positions_B"]["splits"]=[]
					candidates[chrA][chrB][candidate]["positions_B"]["discordants"]=[]
					candidates[chrA][chrB][candidate]["positions_B"]["orientation_contigs"]=[]
					candidates[chrA][chrB][candidate]["positions_B"]["orientation_splits"]=[]
					candidates[chrA][chrB][candidate]["positions_B"]["orientation_discordants"]=[]
					candidates[chrA][chrB][candidate]["positions_B"]["start"]=[]
					candidates[chrA][chrB][candidate]["positions_B"]["end"]=[]

					candidates[chrA][chrB][candidate]["start_B"]=0
					candidates[chrA][chrB][candidate]["end_B"]=0

				if not discordants[chrA][chrB][i][1] in candidates[chrA][chrB][candidate]["samples"]:
					candidates[chrA][chrB][candidate]["sample_discordants"][discordants[chrA][chrB][i][1]]=set([])
					candidates[chrA][chrB][candidate]["sample_splits"][discordants[chrA][chrB][i][1]]=set([])
					candidates[chrA][chrB][candidate]["sample_contigs"][discordants[chrA][chrB][i][1]]=set([])

				candidates[chrA][chrB][candidate]["samples"].add(discordants[chrA][chrB][i][1])

				candidates[chrA][chrB][candidate]["positions_A"]["start"].append(discordants[chrA][chrB][i][8])
				candidates[chrA][chrB][candidate]["positions_A"]["end"].append(discordants[chrA][chrB][i][9])

				candidates[chrA][chrB][candidate]["positions_B"]["start"].append(discordants[chrA][chrB][i][10])
				candidates[chrA][chrB][candidate]["positions_B"]["end"].append(discordants[chrA][chrB][i][11])

				if discordants[chrA][chrB][i][2] == "D":
					candidates[chrA][chrB][candidate]["discordants"].add(discordants[chrA][chrB][i][0])
					candidates[chrA][chrB][candidate]["positions_A"]["discordants"].append(int(discordants[chrA][chrB][i][3]))
					candidates[chrA][chrB][candidate]["positions_A"]["orientation_discordants"].append(discordants[chrA][chrB][i][4])

					candidates[chrA][chrB][candidate]["positions_B"]["discordants"].append(int(discordants[chrA][chrB][i][5]))
					candidates[chrA][chrB][candidate]["positions_B"]["orientation_discordants"].append(discordants[chrA][chrB][i][6])
					candidates[chrA][chrB][candidate]["sample_discordants"][discordants[chrA][chrB][i][1]].add(discordants[chrA][chrB][i][0])

				elif discordants[chrA][chrB][i][2] == "S":
					candidates[chrA][chrB][candidate]["splits"].add(discordants[chrA][chrB][i][0])
					candidates[chrA][chrB][candidate]["positions_A"]["splits"].append(int(discordants[chrA][chrB][i][3]))
					candidates[chrA][chrB][candidate]["positions_A"]["orientation_splits"].append(discordants[chrA][chrB][i][4])

					candidates[chrA][chrB][candidate]["positions_B"]["splits"].append(int(discordants[chrA][chrB][i][5]))
					candidates[chrA][chrB][candidate]["positions_B"]["orientation_splits"].append(discordants[chrA][chrB][i][6])
					candidates[chrA][chrB][candidate]["sample_splits"][discordants[chrA][chrB][i][1]].add(discordants[chrA][chrB][i][0])
				else:
					candidates[chrA][chrB][candidate]["contigs"].add(discordants[chrA][chrB][i][0])
					candidates[chrA][chrB][candidate]["positions_A"]["contigs"].append(int(discordants[chrA][chrB][i][3]))
					candidates[chrA][chrB][candidate]["positions_A"]["orientation_contigs"].append(discordants[chrA][chrB][i][4])

					candidates[chrA][chrB][candidate]["positions_B"]["contigs"].append(int(discordants[chrA][chrB][i][5]))
					candidates[chrA][chrB][candidate]["positions_B"]["orientation_contigs"].append(discordants[chrA][chrB][i][6])
					candidates[chrA][chrB][candidate]["sample_contigs"][discordants[chrA][chrB][i][1]].add(discordants[chrA][chrB][i][0])



	for chrA in candidates:
		for chrB in candidates[chrA]:
			for candidate in candidates[chrA][chrB]:
				candidates[chrA][chrB][candidate]["N_discordants"]=len(candidates[chrA][chrB][candidate]["discordants"])
				candidates[chrA][chrB][candidate]["N_splits"]=len(candidates[chrA][chrB][candidate]["splits"])
				candidates[chrA][chrB][candidate]["N_contigs"]=len(candidates[chrA][chrB][candidate]["contigs"])


				if candidates[chrA][chrB][candidate]["N_splits"] and min_reads <= candidates[chrA][chrB][candidate]["N_splits"]:
					candidates[chrA][chrB][candidate]["posA"]=mode(candidates[chrA][chrB][candidate]["positions_A"]["splits"])
					candidates[chrA][chrB][candidate]["posB"]=mode(candidates[chrA][chrB][candidate]["positions_B"]["splits"])

				elif candidates[chrA][chrB][candidate]["N_contigs"]:
					candidates[chrA][chrB][candidate]["posA"]=mode(candidates[chrA][chrB][candidate]["positions_A"]["contigs"])
					candidates[chrA][chrB][candidate]["posB"]=mode(candidates[chrA][chrB][candidate]["positions_B"]["contigs"])

				elif candidates[chrA][chrB][candidate]["N_splits"]:
					candidates[chrA][chrB][candidate]["posA"]=mode(candidates[chrA][chrB][candidate]["positions_A"]["splits"])
					candidates[chrA][chrB][candidate]["posB"]=mode(candidates[chrA][chrB][candidate]["positions_B"]["splits"])				

				else:
					reverse_A = candidates[chrA][chrB][candidate]["positions_A"]["orientation_discordants"].count("True")
					forward_A = candidates[chrA][chrB][candidate]["positions_A"]["orientation_discordants"].count("False")

					reverse_B = candidates[chrA][chrB][candidate]["positions_B"]["orientation_discordants"].count("True") 
					forward_B = candidates[chrA][chrB][candidate]["positions_B"]["orientation_discordants"].count("False")

					if ( reverse_A >= 5*forward_A or reverse_A*5 <= forward_A ) and ( reverse_B >= 5*forward_B or reverse_B*5 <= forward_B ):
						A_reverse=False
						if reverse_A > forward_A:
							A_reverse=True

						B_reverse=False
						if reverse_B > forward_B:
							B_reverse=True

						if is_mp:
							if A_reverse and not B_reverse:
								candidates[chrA][chrB][candidate]["posA"]=max(candidates[chrA][chrB][candidate]["positions_A"]["discordants"])
								candidates[chrA][chrB][candidate]["posB"]=min(candidates[chrA][chrB][candidate]["positions_B"]["discordants"])

							elif not A_reverse and B_reverse:
								candidates[chrA][chrB][candidate]["posA"]=min(candidates[chrA][chrB][candidate]["positions_A"]["discordants"])
								candidates[chrA][chrB][candidate]["posB"]=max(candidates[chrA][chrB][candidate]["positions_B"]["discordants"])

							elif A_reverse and B_reverse:
								candidates[chrA][chrB][candidate]["posA"]=max(candidates[chrA][chrB][candidate]["positions_A"]["discordants"])
								candidates[chrA][chrB][candidate]["posB"]=max(candidates[chrA][chrB][candidate]["positions_B"]["discordants"])

							else:
								candidates[chrA][chrB][candidate]["posA"]=min(candidates[chrA][chrB][candidate]["positions_A"]["discordants"])
								candidates[chrA][chrB][candidate]["posB"]=min(candidates[chrA][chrB][candidate]["positions_B"]["discordants"])

						else:
							if not A_reverse and B_reverse:
								candidates[chrA][chrB][candidate]["posA"]=max(candidates[chrA][chrB][candidate]["positions_A"]["discordants"])
								candidates[chrA][chrB][candidate]["posB"]=min(candidates[chrA][chrB][candidate]["positions_B"]["discordants"])

							elif A_reverse and not B_reverse:
								candidates[chrA][chrB][candidate]["posA"]=min(candidates[chrA][chrB][candidate]["positions_A"]["discordants"])
								candidates[chrA][chrB][candidate]["posB"]=max(candidates[chrA][chrB][candidate]["positions_B"]["discordants"])

							elif not A_reverse and not B_reverse:
								candidates[chrA][chrB][candidate]["posA"]=max(candidates[chrA][chrB][candidate]["positions_A"]["discordants"])
								candidates[chrA][chrB][candidate]["posB"]=max(candidates[chrA][chrB][candidate]["positions_B"]["discordants"])

							else:
								candidates[chrA][chrB][candidate]["posA"]=min(candidates[chrA][chrB][candidate]["positions_A"]["discordants"])
								candidates[chrA][chrB][candidate]["posB"]=min(candidates[chrA][chrB][candidate]["positions_B"]["discordants"])

					else:
						candidates[chrA][chrB][candidate]["posA"]=mode(candidates[chrA][chrB][candidate]["positions_A"]["discordants"])
						candidates[chrA][chrB][candidate]["posB"]=mode(candidates[chrA][chrB][candidate]["positions_B"]["discordants"])


				candidates[chrA][chrB][candidate]["startB"]=min(candidates[chrA][chrB][candidate]["positions_B"]["start"])
				candidates[chrA][chrB][candidate]["endB"]=max(candidates[chrA][chrB][candidate]["positions_B"]["end"])

				candidates[chrA][chrB][candidate]["startA"]=min(candidates[chrA][chrB][candidate]["positions_A"]["start"])
				candidates[chrA][chrB][candidate]["endA"]=max(candidates[chrA][chrB][candidate]["positions_A"]["end"])

	return(candidates)

#chromosomes=["1","2","3","4","5"]
#samples=["SweGen0001"]
#prefix=sys.argv[1]
#hej=main(prefix,chromosomes,samples,False,150,2)
##print(hej["3"]["5"])
#for entry in hej["3"]["5"]:
#	print(hej["3"]["5"][entry])
#



from setuptools import setup
import numpy

try:
    from Cython.Build import cythonize
    has_cython = True
except ImportError:
    has_cython = False

if has_cython:
    ext_modules = cythonize([
        "tiddit/tiddit_signal.pyx",
        "tiddit/tiddit_coverage.pyx",
        "tiddit/tiddit_cluster.pyx",
        "tiddit/tiddit_coverage_analysis.pyx",
        "tiddit/tiddit_variant.pyx",
        "tiddit/tiddit_contig_analysis.pyx"])
else:
    ext_modules = []

setup(
    name = 'tiddit',
    version = '3.6.1',


    url = "https://github.com/SciLifeLab/TIDDIT",
    author = "Jesper Eisfeldt",
    author_email= "jesper.eisfeldt@scilifelab.se",
    ext_modules = ext_modules,
    include_dirs=[numpy.get_include()],
    packages = ['tiddit'],
    install_requires = ['numpy','pysam'],
    entry_points = {'console_scripts': ['tiddit = tiddit.__main__:main']},

)



import pysam
import sys

def main(bam_header,library,sample_id,version):

	vcf_header=[]	

	vcf_header.append("##fileformat=VCFv4.1")
	vcf_header.append("##source=TIDDIT-" + version)

        #declare the events classified by TIDDIT

	vcf_header.append("##ALT=<ID=DEL,Description=\"Deletion\">")
	vcf_header.append("##ALT=<ID=DUP,Description=\"Duplication\">")
	vcf_header.append("##ALT=<ID=DUP:TANDEM,Description=\"Tandem duplication\">")
	vcf_header.append("##ALT=<ID=DUP:INV,Description=\"Inverted tandem duplication\">")
	vcf_header.append("##ALT=<ID=INV,Description=\"Inversion\">")
	vcf_header.append("##ALT=<ID=INS,Description=\"Insertion\">")
	vcf_header.append("##ALT=<ID=BND,Description=\"Break end\">")

	#print chromosomes and length
	for contig in bam_header["SQ"]:
		#print(contig)
		vcf_header.append("##contig=<ID={},length={}>".format(contig["SN"],contig["LN"]) )

	#declare the info field

	vcf_header.append("##INFO=<ID=SVTYPE,Number=1,Type=String,Description=\"Type of structural variant\">")
	vcf_header.append("##INFO=<ID=END,Number=1,Type=Integer,Description=\"End of an intra-chromosomal variant\">")
	vcf_header.append("##INFO=<ID=SVLEN,Number=.,Type=Integer,Description=\"Difference in length between REF and ALT alleles\">")
	vcf_header.append("##INFO=<ID=LFA,Number=2,Type=Integer,Description=\"Read-pairs and split reads in region A\">")
	vcf_header.append("##INFO=<ID=LFB,Number=2,Type=Integer,Description=\"Read-pairs and split reads in region B\">")
	vcf_header.append("##INFO=<ID=LTE,Number=2,Type=Integer,Description=\"Read-pairs and split reads supporting the event\">")
	vcf_header.append("##INFO=<ID=CTG,Number=1,Type=String,Description=\"Sequence of contig\">")
	vcf_header.append("##INFO=<ID=REGIONA,Number=2,Type=Integer,Description=\"Start and end of regionB\">")
	vcf_header.append("##INFO=<ID=REGIONB,Number=2,Type=Integer,Description=\"Start and end of regionB\">")

	#Declare the filters

	vcf_header.append("##FILTER=<ID=BelowExpectedLinks,Description=\"The number of links or reads between A and B is too small\">")
	vcf_header.append("##FILTER=<ID=FewLinks,Description=\"Unexpectedly low fraction of discordant reads betwen A and B\">")
	vcf_header.append("##FILTER=<ID=UnexpectedCoverage,Description=\"The coverage of the window on chromosome B or A is higher than 4*average coverage\">")
	vcf_header.append("##FILTER=<ID=Smear,Description=\"Window A and Window B overlap\">")
	vcf_header.append("##FILTER=<ID=RegionalQ,Description=\"The mapping quality of the region is lower than the user set limit\">")
	vcf_header.append("##FILTER=<ID=MinSize,Description=\"The variant is smaller than the user set limit\">")
	vcf_header.append("##FILTER=<ID=Ploidy,Description=\"Intrachromosomal variant on a chromosome having 0 ploidy\">")
	vcf_header.append("##FILTER=<ID=SplitsVSDiscs,Description=\"large variant supported mainly by split reads (and not discorant pairs) \">")
	vcf_header.append("##FILTER=<ID=Density,Description=\"The discordant reads cluster too tightly\">")

	#set format

	vcf_header.append("##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">")
	vcf_header.append("##FORMAT=<ID=CN,Number=1,Type=Integer,Description=\"Copy number genotype for imprecise events\">")
	vcf_header.append("##FORMAT=<ID=DV,Number=1,Type=Integer,Description=\"Number of paired-ends that support the event\">")
	vcf_header.append("##FORMAT=<ID=RV,Number=1,Type=Integer,Description=\"Number of split reads that support the event\">")
	vcf_header.append("##FORMAT=<ID=DR,Number=2,Type=Integer,Description=\"Number of paired-ends that supporting the reference allele (breakpoint A, and B)\">")
	vcf_header.append("##FORMAT=<ID=RR,Number=2,Type=Integer,Description=\"Number of reads supporting the reference allele (breakpoint A, and B)\">")
	vcf_header.append("##FORMAT=<ID=COV,Number=3,Type=Float,Description=\"Coverage (at A,B, and between)\">")
	vcf_header.append("##FORMAT=<ID=LQ,Number=2,Type=Float,Description=\"Fraction of low quality reads\">")

	#library statistics line
	vcf_header.append("##LibraryStats=TIDDIT-{} Coverage={}  ReadLength={} MeanInsertSize={} STDInsertSize={} Reverse_Forward={}".format(version,library["avg_coverage"],library["avg_read_length"],library["avg_insert_size"],library["std_insert_size"],library["mp"] ) ) 

	#command used to launch tiddit
	vcf_header.append("##TIDDITcmd=\"" + " ".join(sys.argv) + "\"")

	vcf_header.append("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t"+sample_id)
	return("\n".join(vcf_header))


#generate test header
#bam_file_name=sys.argv[1]

#samfile = pysam.AlignmentFile(bam_file_name, "r")
#bam_header=samfile.header
#samfile.close()

#try:
#	sample_id=header["RG"][0]["SM"]
#
#except:
#	sample_id=bam_file_name.split("/")[-1].split(".")[0]

#library={}
#version="4.0.0"

#library["avg_read_length"]=151
#library["avg_insert_size"]=350
#library["std_insert_size"]=400
#library["mp"]=True
#library["avg_coverage"]=35


#print(main(bam_header,library,sample_id,version))

